#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.cdc import MultiReg

# Core <-> PHY Layouts -----------------------------------------------------------------------------

"""
Stream layout for LiteSPICore->PHY connection:
data - flash byte address when cmd=1, data to transmit when cmd=2, unused in cmd=0
len - xfer length (in bits)
width - xfer width (1/2/4/8)
mask - dq output enable control (1 enables a output on a particular pin)
"""
spi_core2phy_layout = [
    ("data", 32),
    ("len",   6),
    ("width", 4),
    ("mask",  8),
]
"""
Stream layout for PHY->LiteSPICore connection
data - 32-bits of data from flash
"""
spi_phy2core_layout = [
    ("data", 32),
]

MMAP_DEFAULT_TIMEOUT = 256


# Helpers ------------------------------------------------------------------------------------------

class ResyncReg(Module):
    def __init__(self, src, dst, clock_domain):
        if clock_domain == "sys":
            self.comb += dst.eq(src)
        else:
            self.specials += MultiReg(src, dst, clock_domain)



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause


class MetaSizes(type):
    """Meta Sizes

    The ``MetaSizes`` class makes sure that all provided size parameters
    in chip are consistent which are:
        * total size
        * size per unit
        * number of units
    """

    def _table(self, mod_name, name):
        size_name  = "{}_size".format(name)
        total_name = "total_{}s".format(name)
        ts_name    = "total_size"

        provided_size_per_unit   = getattr(self, "provided_"+size_name)
        provided_number_of_units = getattr(self, "provided_"+total_name)
        provided_total_size      = getattr(self, "provided_total_size")

        size_per_unit   = getattr(self, size_name)
        number_of_units = getattr(self, total_name)
        total_size = self.total_size

        calc_size_per_unit   = int(total_size / number_of_units)
        calc_number_of_units = int(total_size / size_per_unit)
        calc_total_size      = int(size_per_unit * number_of_units)

        def check(a, b):
            if a is not None and a != b:
                return '!='
            return '=='

        def si(a, f=str):
            if a is None:
                return ''
            else:
                return f(int(a))

        return """
| {mod_name:14s}                 |     Provided == Calculated   |     Hex    |
|-----------------|--------------|------------------------------|------------|
|   size_per_unit | {size_name_:>12s} | {p_size_:>12s} {size__check} {c_size_:<12s} | {h_size_:>10s} | {p_tsize:>10s} / {p_units:10s} ({ts_name___:>10s} / {total_name:10s})
| number_of_units | {total_name:>12s} | {p_units:>12s} {units_check} {c_units:<12s} | {h_units:>10s} | {p_tsize:>10s} / {p_size_:10s} ({ts_name___:>10s} / {size_name_:10s})
|      total_size | {ts_name___:>12s} | {p_tsize:>12s} {tsize_check} {c_tsize:<12s} | {h_tsize:>10s} | {p_size_:>10s} * {p_units:10s} ({size_name_:>10s} * {ts_name___:10s})
""".format(
    mod_name = mod_name,
    name = name,
    size_name_ = size_name,
    total_name = total_name,
    ts_name___ = ts_name,
    size__check = check(provided_size_per_unit,   calc_size_per_unit  ),
    units_check = check(provided_number_of_units, calc_number_of_units),
    tsize_check = check(provided_total_size,      calc_total_size     ),
    p_size_ = si(provided_size_per_unit),
    p_units = si(provided_number_of_units),
    p_tsize = si(provided_total_size),
    c_size_ = si(calc_size_per_unit),
    c_units = si(calc_number_of_units),
    c_tsize = si(calc_total_size),
    h_size_ = si(calc_size_per_unit, hex),
    h_units = si(calc_number_of_units, hex),
    h_tsize = si(calc_total_size, hex),
)

    def _fix_and_check_sized(self, mod_name, name):
        """Makes sure that XXX_size, XXX_total and total_size match.
        IE If XXX == pages then, page_size * total_pages == total_size
        """
        size_name  = "{}_size".format(name)
        total_name = "total_{}s".format(name)

        provided_size_per_unit   = getattr(self, size_name,    None)
        provided_number_of_units = getattr(self, total_name,   None)
        provided_total_size      = getattr(self, "total_size", None)
        setattr(self, "provided_"+size_name,   provided_size_per_unit)
        setattr(self, "provided_"+total_name,  provided_number_of_units)

        provided = 0
        if provided_size_per_unit is not None:
            provided += 1
        if provided_number_of_units is not None:
            provided += 1
        if provided_total_size is not None:
            provided += 1

        assert provided >= 2, """\
{mod_name}: Must provide two of {name}_size={} (size_per_unit), total_{name}s={} (number_of_units), total_size={}
""".format(provided_size_per_unit, provided_number_of_units, provided_total_size, mod_name=mod_name, name=name)

        if provided_size_per_unit is None:
            size_per_unit = int(provided_total_size / provided_number_of_units)
        else:
            size_per_unit = int(provided_size_per_unit)

        if provided_number_of_units is None:
            number_of_units = int(provided_total_size / provided_size_per_unit)
        else:
            number_of_units = int(provided_number_of_units)

        if provided_total_size is None:
            total_size = int(provided_size_per_unit * provided_number_of_units)
        else:
            total_size = int(provided_total_size)

        setattr(self, size_name,  size_per_unit)
        setattr(self, total_name, number_of_units)

        assert   size_per_unit ==    total_size / number_of_units, self._table(mod_name, name)
        assert number_of_units ==    total_size / size_per_unit,   self._table(mod_name, name)
        assert      total_size == size_per_unit * number_of_units, self._table(mod_name, name)
        if provided_total_size is not None:
            assert provided_total_size == total_size, """\
{mod_name}: Tried to set size to {} but size was already set to {}
""".format(provided_total_size, self.total_size, mod_name=mod_name)

    def __init__(self, name, *args, **kw):
        if name == "SpiNorFlashModule":
           return
        setattr(self, "provided_total_size", getattr(self, "total_size", None))
        self._fix_and_check_sized(name, "page")
        #self._fix_and_check_sized(name, "sector")
        self.total_size = self.page_size * self.total_pages
        self.table = lambda x: self._table(name, x)




#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.cdc import MultiReg

from litex.gen.genlib.misc import WaitTimer

from litespi.common import *
from litespi.clkgen import DDRLiteSPIClkGen

from litex.soc.interconnect import stream
from litex.soc.interconnect.csr import *

from litex.build.io import DDRTristate

from litex.soc.integration.doc import AutoDoc


# LiteSPI DDR PHY Core -----------------------------------------------------------------------------

class LiteSPIDDRPHYCore(Module, AutoCSR, AutoDoc):
    """LiteSPI PHY DDR instantiator

    The ``DDRLiteSPIPHYCore`` class provides a generic PHY that can be connected to the ``LiteSPICore``.

    It supports single/dual/quad/octal output reads from the flash chips.

    You can use this class only with devices that supports the DDR primitives.

    The following diagram shows how each clock configuration option relates to outputs and input sampling in DDR mode:

    .. wavedrom:: ../../doc/ddr-timing-diagram.json

    Parameters
    ----------
    pads : Object
        SPI pads description.

    flash : SpiNorFlashModule
        SpiNorFlashModule configuration object.

    Attributes
    ----------
    source : Endpoint(spi_phy2core_layout), out
        Data stream.

    sink : Endpoint(spi_core2phy_layout), in
        Control stream.

    cs : Signal(), in
        Flash CS signal.
    """
    def __init__(self, pads, flash, cs_delay, extra_latency=0):
        self.source = source = stream.Endpoint(spi_phy2core_layout)
        self.sink   = sink   = stream.Endpoint(spi_core2phy_layout)
        self.cs     = Signal()

        if hasattr(pads, "miso"):
            bus_width = 1
            pads.dq   = [pads.mosi, pads.miso]
        else:
            bus_width = len(pads.dq)

        assert bus_width in [1, 2, 4, 8]

        # Check if number of pads matches configured mode.
        assert flash.check_bus_width(bus_width)

        self.addr_bits  = addr_bits  = flash.addr_bits
        self.ddr        = ddr        = flash.ddr

        assert not ddr

        # Clock Generator.
        self.submodules.clkgen = clkgen = DDRLiteSPIClkGen(pads)

        # CS control.
        cs_timer  = WaitTimer(cs_delay + 1) # Ensure cs_delay cycles between XFers.
        cs_enable = Signal()
        self.submodules += cs_timer
        self.comb += cs_timer.wait.eq(self.cs)
        self.comb += cs_enable.eq(cs_timer.done)
        self.comb += pads.cs_n.eq(~cs_enable)

        # I/Os.
        data_bits = 32

        dq_o  = Array([Signal(len(pads.dq)) for _ in range(2)])
        dq_i  = Array([Signal(len(pads.dq)) for _ in range(2)])
        dq_oe = Array([Signal(len(pads.dq)) for _ in range(2)])

        for i in range(len(pads.dq)):
            self.specials += DDRTristate(
                io  = pads.dq[i],
                o1  =  dq_o[0][i],  o2 =  dq_o[1][i],
                oe1 = dq_oe[0][i], oe2 = dq_oe[1][i],
                i1  =  dq_i[0][i],  i2 =  dq_i[1][i]
            )

        # Data Shift Registers.
        sr_cnt       = Signal(8, reset_less=True)
        sr_out_load  = Signal()
        sr_out_shift = Signal()
        sr_out       = Signal(len(sink.data), reset_less=True)
        sr_in_shift  = Signal()
        sr_in        = Signal(len(sink.data), reset_less=True)

        # Data Out Shift.
        self.comb += [
            dq_oe[1].eq(sink.mask),
            Case(sink.width, {
                1:  dq_o[1].eq(sr_out[-1:]),
                2:  dq_o[1].eq(sr_out[-2:]),
                4:  dq_o[1].eq(sr_out[-4:]),
                8:  dq_o[1].eq(sr_out[-8:]),
            })
        ]
        self.sync += If(sr_out_load,
            sr_out.eq(sink.data << (len(sink.data) - sink.len))
        )
        self.sync += If(sr_out_shift,
            dq_oe[0].eq(dq_oe[1]),
            dq_o[0].eq(dq_o[1]),
            Case(sink.width, {
                1 : sr_out.eq(Cat(Signal(1), sr_out)),
                2 : sr_out.eq(Cat(Signal(2), sr_out)),
                4 : sr_out.eq(Cat(Signal(4), sr_out)),
                8 : sr_out.eq(Cat(Signal(8), sr_out)),
            })
        )

        # Data In Shift.
        self.sync += If(sr_in_shift,
            Case(sink.width, {
                1 : sr_in.eq(Cat(dq_i[0][1],  sr_in)), # 1: pads.miso
                2 : sr_in.eq(Cat(dq_i[0][:2], sr_in)),
                4 : sr_in.eq(Cat(dq_i[0][:4], sr_in)),
                8 : sr_in.eq(Cat(dq_i[0][:8], sr_in)),
            })
        )

        # FSM.
        self.submodules.fsm = fsm = FSM(reset_state="WAIT-CMD-DATA")
        fsm.act("WAIT-CMD-DATA",
            # Stop Clk.
            NextValue(clkgen.en, 0),
            # Wait for CS and a CMD from the Core.
            If(cs_enable & sink.valid,
                # Load Shift Register Count/Data Out.
                NextValue(sr_cnt, sink.len - sink.width),
                sr_out_load.eq(1),
                # Start XFER.
                NextState("XFER")
            )
        )

        fsm.act("XFER",
            # Generate Clk.
            NextValue(clkgen.en, 1),

            # Data In Shift.
            sr_in_shift.eq(1),

            # Data Out Shift.
            sr_out_shift.eq(1),

            # Shift Register Count Update/Check.
            NextValue(sr_cnt, sr_cnt - sink.width),
            # End XFer.
            If(sr_cnt == 0,
                NextValue(sr_cnt, (2 + 2*extra_latency)*sink.width), # FIXME: Explain magic numbers.
                NextState("XFER-END"),
            ),
        )
        fsm.act("XFER-END",
            # Stop Clk.
            NextValue(clkgen.en, 0),

            # Data In Shift.
            sr_in_shift.eq(1),

            # Shift Register Count Update/Check.
            NextValue(sr_cnt, sr_cnt - sink.width),
            If(sr_cnt == 0,
                sink.ready.eq(1),
                NextState("SEND-STATUS-DATA"),
            ),
        )
        self.comb += source.data.eq(sr_in)
        fsm.act("SEND-STATUS-DATA",
            # Send Data In to Core and return to WAIT when accepted.
            source.valid.eq(1),
            source.last.eq(1),
            If(source.ready,
                NextState("WAIT-CMD-DATA"),
            )
        )



#!/usr/bin/env python3

#
# This file is part of LiteSPI.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2022 Victor Suarez Rovere <suarezvictor@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

"""
LiteSPI standalone core generator

LiteSPI aims to be directly used as a python package when the SoC is created using LiteX. However,
for some use cases it could be interesting to generate a standalone verilog file of the core:
- integration of the core in a SoC using a more traditional flow.
- need to version/package the core.
- avoid Migen/LiteX dependencies.
- etc...

The standalone core is generated from a YAML configuration file that allows the user to generate
easily a custom configuration of the core.

"""

import os
import yaml
import argparse
import inspect

from migen import *

from litex.build.generic_platform import *
from litex.build.sim import SimPlatform
from litex.build.xilinx.platform import XilinxPlatform
from litex.build.altera.platform import AlteraPlatform
from litex.build.lattice.platform import LatticePlatform

from litex.soc.interconnect import wishbone
from litex.soc.interconnect import axi
from litex.soc.integration.soc_core import *
from litex.soc.integration.builder import *
from litex.soc.integration.soc import SoCRegion

import litespi.modules
from litespi.spi_nor_flash_module import SpiNorFlashModule
from litespi.opcodes import SpiNorFlashOpCodes as Codes

from litespi import LiteSPI
from litespi.phy.generic import LiteSPIPHY

# IOs ----------------------------------------------------------------------------------------------

_io = [
    # Clk / Rst
    ("clk", 0, Pins(1)),
    ("rst", 1, Pins(1)),

    # SPI-Flash (X1).
    ("spiflash", 0,
        Subsignal("cs_n", Pins(1)),
        Subsignal("clk",  Pins(1)),
        Subsignal("mosi", Pins(1)),
        Subsignal("miso", Pins(1)),
        Subsignal("wp",   Pins(1)),
        Subsignal("hold", Pins(1)),
    ),

    # SPI-Flash (X4).
    ("spiflash4x", 0,
        Subsignal("cs_n", Pins(1)),
        Subsignal("clk",  Pins(1)),
        Subsignal("dq",   Pins(4)),
    ),
]

# LiteSPI Core -------------------------------------------------------------------------------------

class LiteSPICore(SoCMini):
    def __init__(self, platform, module, mode="x4", rate="1:1", divisor="1",
        bus_standard   = "wishbone",
        bus_endianness = "big",
        with_master    = False,
        sim            = False
    ):
        # CRG --------------------------------------------------------------------------------------
        self.submodules.crg = CRG(platform.request("clk"), platform.request("rst"))

        # SoCMini ----------------------------------------------------------------------------------
        SoCMini.__init__(self, platform, clk_freq=int(1e6))

        # SPI Flash Module -------------------------------------------------------------------------

        # Get available modules.
        modules_cls = inspect.getmembers(litespi.modules, inspect.isclass)
        modules     = {name: cls for name, cls in modules_cls if issubclass(cls, SpiNorFlashModule)}

        # Check that selected module is supported.
        def print_supported_modules():
            r = ""
            i = 0
            modules_per_line = 8
            modules_interval = 24
            for name, module in modules.items():
                r += f"{name}{' '*(modules_interval-len(name))}"
                if (i%modules_per_line) == (modules_per_line-1):
                    r += "\n"
                i += 1
            return r

        if not module in modules:
            raise ValueError(f"Unsupported SPI module {module}, supported:\n{print_supported_modules()}")

        # Check that selected width is supported in module.
        assert mode in ["x1", "x4"]
        module_width = {
          "x1" : 1,
          "x4" : 4,
        }[mode]
        spiflash_module_cls = modules[module]
        spiflash_module     = spiflash_module_cls(getattr(Codes, f"READ_1_1_{module_width}"))

        # SPI Flash PHY ----------------------------------------------------------------------------

        if sim:
            from litespi.phy.model import LiteSPIPHYModel
            spiflash_phy = LiteSPIPHYModel(spiflash_module, init=[i for i in range(16)]) # FIXME: Allow custom init?
            self.submodules += spiflash_phy
        else:
            pads = self.platform.request("spiflash" if mode == "x1" else "spiflash4x")
            spiflash_phy = LiteSPIPHY(
                pads            = pads,
                flash           = spiflash_module,
                device          = platform.device,
                default_divisor = int(divisor),
                rate            = rate
            )
            self.submodules += spiflash_phy


        # SPI Flash Core / MMAP --------------------------------------------------------------------

        assert bus_standard in ["wishbone", "axi-lite"]

        spiflash_core = LiteSPI(
            phy             = spiflash_phy,
            mmap_endianness = bus_endianness,
            with_master     = with_master,
            with_mmap       = True,
            with_csr        = False
        )
        self.submodules += spiflash_core

        # Wishbone.
        if bus_standard == "wishbone":
            # LiteSPI is already in Wishbone, just expose the Bus.
            platform.add_extension(spiflash_core.bus.get_ios("bus"))
            self.comb += spiflash_core.bus.connect_to_pads(platform.request("bus"), mode="slave")

            # Expose Ctrl Bus.
            if with_master:
                master_bus = wishbone.Interface(address_width=32, data_width=32)
                platform.add_extension(master_bus.get_ios("master"))
                self.comb += master_bus.connect_to_pads(self.platform.request("master"), mode="slave")
                self.bus.add_master(master=master_bus)

        # AXI-Lite.
        if bus_standard == "axi-lite":
            # LiteSPI is in Wishbone, converter to AXI-Lite and expose the AXI-Lite Bus.
            axil_bus = axi.AXILiteInterface(address_width=32, data_width=32)
            platform.add_extension(axil_bus.get_ios("bus"))
            self.submodules += axi.AXILite2Wishbone(axil_bus, spiflash_core.bus)
            self.comb += axil_bus.connect_to_pads(platform.request("bus"), mode="slave")

            # Expose Ctrl Bus.
            if with_master:
                master_bus = axi.AXILiteInterface(address_width=32, data_width=32)
                platform.add_extension(master_bus.get_ios("master"))
                self.comb += master_bus.connect_to_pads(self.platform.request("master"), mode="slave")
                self.bus.add_master(master=master_bus)

# Build --------------------------------------------------------------------------------------------

def main():
    parser = argparse.ArgumentParser(description="LiteSPI standalone core generator.")
    parser.add_argument("--clk-freq",       default="100e6",      help="Input Clk Frequency.")
    parser.add_argument("--vendor",         default="xilinx",     help="FPGA Vendor.")
    parser.add_argument("--module",         default="S25FL128L",  help="SPI Flash Module.")
    parser.add_argument("--mode",           default="x1",         help="SPI Mode (x1, x4).")
    parser.add_argument("--rate",           default="1:1",        help="SPI Flash Core rate (1:1, 1:2).")
    parser.add_argument("--divisor",        default="1",          help="SPI PHY Clk Divisor.")
    parser.add_argument("--bus-standard",   default="wishbone",   help="Bus Standard (wishbone, axi-lite).")
    parser.add_argument("--bus-endianness", default="big",        help="Bus Endianness (big, little).")
    parser.add_argument("--with-master",    action='store_true',  help="Enable Master (For Control & Writes).")
    parser.add_argument("--sim",            action='store_true',  help="Integrate LiteSPIPHYModel in core for simulation.")
    args = parser.parse_args()

    # Generate core --------------------------------------------------------------------------------
    platform_cls = {
        "xilinx"  : XilinxPlatform,
        "altera"  : AlteraPlatform,
        "intel"   : AlteraPlatform,
        "lattice" : LatticePlatform
    }[args.vendor]
    platform = platform_cls(device="", io=_io)
    core     = LiteSPICore(platform,
        module         = args.module,
        mode           = args.mode,
        rate           = args.rate,
        divisor        = args.divisor,
        bus_standard   = args.bus_standard,
        bus_endianness = args.bus_endianness,
        with_master    = args.with_master,
        sim            = args.sim,
    )
    builder  = Builder(core, output_dir="build")
    builder.build(build_name="litespi_core", run=False)

if __name__ == "__main__":
    main()



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.fsm import FSM, NextState

from litex.soc.interconnect import stream

from litespi.common import *

class LiteSPIPHYModelCore(Module):
    def __init__(self, flash, init=None, latency=4):
        self.source     = source = stream.Endpoint(spi_phy2core_layout)
        self.sink       = sink   = stream.Endpoint(spi_core2phy_layout)
        self.dummy_bits = Signal(8)
        self.cs         = Signal()

        mem         = Memory(32, flash.total_size//4, init=init)
        read_addr   = Signal(32)

        read_port   = mem.get_port(async_read=True)

        self.specials += mem, read_port

        self.comb   += read_port.adr.eq(read_addr)

        cmd_bits    = 8

        delay_cnt   = Signal(8)
        xfer_len    = Signal(8)
        xfer_width  = Signal(8)

        def delay_gen(next_state, width=xfer_width, bits=xfer_len, op=[]):
            res = [
                If(delay_cnt == (bits - width + latency*width),
                    source.valid.eq(1),
                    If(source.ready,
                        NextValue(delay_cnt, 0),
                        NextState(next_state),
                        *op,
                    ),
                ).Else(
                    NextValue(delay_cnt, delay_cnt + width),
                )
            ]

            return res

        def wait_gen(next_state, op=[]):
            res = [
                sink.ready.eq(1),
                If(sink.valid,
                    NextValue(xfer_len, sink.len),
                    NextValue(xfer_width, sink.width),
                    NextState(next_state),
                    *op,
                )
            ]

            return res

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            sink.ready.eq(self.cs),
            NextValue(read_addr, 0),
            NextValue(delay_cnt, 0),
            If(sink.valid & sink.ready,
                NextValue(xfer_len, sink.len),
                NextValue(xfer_width, sink.width),
                If((sink.len == cmd_bits) & (sink.data == flash.read_opcode.code),
                    NextState("CMD-DELAY"),
                ).Else(
                    NextState("OTHER-DELAY"),
                )
            )
        )

        fsm.act("OTHER-DELAY",
            delay_gen("OTHER-WAIT"),
            If(~self.cs,
                NextState("IDLE")
            )
        )
        fsm.act("OTHER-WAIT",
            wait_gen("OTHER-DELAY"),
            If(~self.cs,
                NextState("IDLE")
            )
        )

        fsm.act("CMD-DELAY",
            delay_gen("ADDR-WAIT")
        )
        fsm.act("ADDR-WAIT",
            wait_gen("ADDR-DELAY", op=[NextValue(read_addr, sink.data[2:])]),
        )
        fsm.act("ADDR-DELAY",
            If(self.dummy_bits != 0, 
                delay_gen("DUMMY-WAIT")
            ).Else(
                delay_gen("DATA-WAIT")
            )
        )
        fsm.act("DUMMY-WAIT",
            wait_gen("DUMMY-DELAY"),
        )
        fsm.act("DUMMY-DELAY",
            delay_gen("DATA-WAIT"),
        )
        fsm.act("DATA-WAIT",
            If(~self.cs,
                NextState("IDLE")
            ),
            wait_gen("DATA-DELAY"),
        )
        fsm.act("DATA-DELAY",
            If(~self.cs,
                NextState("IDLE")
            ),
            source.data.eq(read_port.dat_r),
            delay_gen("DATA-WAIT", op=[NextValue(read_addr, read_addr+1)]),
        )

class LiteSPIPHYModel(Module):
    """LiteSPI PHY simulation model.

    The ``LiteSPIPHYModel`` class provides a simulation model that can be used instead of the real PHY.

    init : []
        Initialization data for the flash contents.

    Attributes
    ----------
    source : Endpoint(spi_phy2core_layout), out
        Data stream.

    sink : Endpoint(spi_core2phy_layout), in
        Control stream.

    cs : Signal()
        Dummy flash CS signal.
    """
    def __init__(self, flash, clock_domain="sys", init=None):
        self.phy        = LiteSPIPHYModelCore(flash, init=init)
        self.dummy_bits = self.phy.dummy_bits
        self.source     = self.phy.source
        self.sink       = self.phy.sink
        self.cs         = self.phy.cs
        self.flash      = flash

        # # #

        if clock_domain != "sys":
            self.clock_domains.cd_litespi = ClockDomain()
            self.phy = ClockDomainsRenamer("litespi")(self.phy)
            self.comb += self.cd_litespi.clk.eq(ClockSignal(clock_domain))
            self.comb += self.cd_litespi.rst.eq(ResetSignal(clock_domain))

        self.submodules.spiflash_phy = self.phy



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from litespi.spi_nor_flash_module import SpiNorFlashModule
from litespi.opcodes import SpiNorFlashOpCodes as Codes
from litespi.ids import SpiNorFlashManufacturerIDs


class TestFlashModule(unittest.TestCase):
    class GoodDummyChip(SpiNorFlashModule):

        manufacturer_id = SpiNorFlashManufacturerIDs.NONJEDEC
        device_id = 0x0204
        name = "dummychip1"

        total_size  =    2097152   # bytes
        page_size   =        256   # bytes
        total_pages =       8192

        supported_opcodes = [
            Codes.READ_1_1_1,
            Codes.READ_1_1_1_FAST,
            Codes.READ_1_1_2_4B,
            Codes.READ_1_4_4_DTR_4B,
            Codes.READ_1_1_8,
            Codes.PP_1_1_1,
        ]
        dummy_bits = 8


    def test_spi_nor_flash_module_configuration(self):
        # Configure to slow mode
        chip = self.GoodDummyChip(Codes.READ_1_1_1)
        self.assertEqual(chip.cmd_width, 1)
        self.assertEqual(chip.addr_width, 1)
        self.assertEqual(chip.bus_width, 1)
        self.assertEqual(chip.addr_bits, 24)
        self.assertEqual(chip.fast_mode, False)
        self.assertEqual(chip.ddr, False)

        # Reconfigure to fast mode
        chip.read_opcode = Codes.READ_1_1_1_FAST
        self.assertEqual(chip.cmd_width, 1)
        self.assertEqual(chip.addr_width, 1)
        self.assertEqual(chip.bus_width, 1)
        self.assertEqual(chip.addr_bits, 24)
        self.assertEqual(chip.fast_mode, True)
        self.assertEqual(chip.ddr, False)

        # Reconfigure to dual, 32-bit address
        chip.read_opcode = Codes.READ_1_1_2_4B
        self.assertEqual(chip.cmd_width, 1)
        self.assertEqual(chip.addr_width, 1)
        self.assertEqual(chip.bus_width, 2)
        self.assertEqual(chip.addr_bits, 32)
        self.assertEqual(chip.fast_mode, True)
        self.assertEqual(chip.ddr, False)

        # Reconfigure to quad, addr_width=4, 32-bit address with DTR
        chip.read_opcode = Codes.READ_1_4_4_DTR_4B
        self.assertEqual(chip.cmd_width, 1)
        self.assertEqual(chip.addr_width, 4)
        self.assertEqual(chip.bus_width, 4)
        self.assertEqual(chip.addr_bits, 32)
        self.assertEqual(chip.fast_mode, True)
        self.assertEqual(chip.ddr, True)

        # Reconfigure to octal
        chip.read_opcode = Codes.READ_1_1_8
        self.assertEqual(chip.cmd_width, 1)
        self.assertEqual(chip.addr_width, 1)
        self.assertEqual(chip.bus_width, 8)
        self.assertEqual(chip.addr_bits, 24)
        self.assertEqual(chip.fast_mode, True)
        self.assertEqual(chip.ddr, False)

        # Nonsupported mode
        with self.assertRaises(ValueError):
            chip = self.GoodDummyChip(Codes.READ_1_8_8)

    def test_meta_sizes(self):
        with self.assertRaises(AssertionError):
            class BadSizesChip(SpiNorFlashModule):

                manufacturer_id = SpiNorFlashManufacturerIDs.NONJEDEC
                device_id = 0x21ab
                name = "dummychip2"

                total_size  =    2097152   # bytes
                page_size   =        256   # bytes
                total_pages =       8191   # inconsistent total_pages

                supported_opcodes = [
                    Codes.READ_1_1_1,
                    Codes.PP_1_1_1,
                ]

                dummy_bits = 8

    def test_spi_nor_flash_module_missing_attr(self):
        class MissingAttrChip(SpiNorFlashModule):

            manufacturer_id = SpiNorFlashManufacturerIDs.NONJEDEC
            device_id = 0x37cd
            name = "dummychip3"

            total_size  =    2097152   # bytes
            page_size   =        256   # bytes
            total_pages =       8192

            supported_modes = [
                Codes.READ_1_1_1,
                Codes.PP_1_1_1,
            ]

            # missing dummy_bits

        with self.assertRaises(AssertionError):
            chip = MissingAttrChip(Codes.READ_1_1_1)

